#ifndef GDBSCAN_H
#define GDBSCAN_H

#include "linlist.h"
#include <limits.h>
#include <float.h>

#ifdef __BORLANDC__
#include <iostream.h>
#include <strstrea.h>
#include <math.h>
#include <stdlib.h>
#else
#include <iostream.h>
#include <backward/strstream>
#include <stream.h>
#include <math.h>
#include <stdlib.h>
#endif

#define DataList LinList<DATA>
typedef long LONGINT;
typedef LONGINT CLUSTERID;

const int ID_NOISE 		 =  0;
const int ID_UNCLASSIFIED = -1;

// ********** Class GDBSCAN_DATA ***********************************************
class GDBSCAN_DATA
{
public:
   LONGINT     id;                                             // Object Id
   CLUSTERID   clusterId;
   virtual int compare(GDBSCAN_DATA *dt) = 0; 						// see GDBSCAN_newSeedPoint..
};

// ********** Class clusterInfo ************************************************
template <class DATA>
class clusterInfo
{
public:
	// abstract virtual functions that must be redeclared to create clusterInfo
	virtual void clear()=0;													// new clusters
   virtual void createTemp(CLUSTERID Id)=0;
	virtual void updateTemp(DATA *d)=0;									// called after _set_clusterId(p, ID_..)
   virtual void commitTemp()=0;

	// virtual functions / optional hook functions
  	virtual void showNoise(DATA *o) {} 									// called after _set_clusterId(p, ID_NOISE)
};

// ********** Class GDBSCAN_database *******************************************
template <class DATA>
class GDBSCAN_database
{
public:
	GDBSCAN_database() {}
	virtual ~GDBSCAN_database() {}

	// abstract virtual functions that must be redeclared to access the data
	virtual void 	   GDBSCAN_cursor()=0;					  			// open cursor for linear scan of database
	virtual DATA 	  *GDBSCAN_fetch()=0;					  			// get next element (copy, allocate new object)
	virtual void 	  	GDBSCAN_set_clusterId(DATA	*o, 		 	 	// update objetct o (may contain Id or other kind of reference)
   											  CLUSTERID Id)=0;			//		i.e. set clusterId of object o to Id

	// abstract virtual functions that must be redeclared to implement NPred and wCard
	virtual void 		GDBSCAN_retrieveNeighborhood(DATA *o, DataList **result)=0;
	/*	retrieves neighborhood according to desired NPred
   	result must be of the DataList type which is a LinList<DATA>
      result must be CREATED in _retrieveNeighborhood (*result=new ..).
      Therefore it can be a derived class of LinList<DATA> - like SortedLinList<DATA>
   */

	// virtual functions / optional hook functions
	virtual void 		GDBSCAN_prepareDatabase();						// optional hook function, to prepare clustering

	// wCard function
	virtual float 		GDBSCAN_wCard(DATA *o); 	               // weight/select single object (DATA) for speed (no DataList needed)
	virtual float 		GDBSCAN_wCard(DataList *result)=0; 			// weight for set of objects

   // wCard specific parameters
	float					GDBSCAN_minCard;									// threshold value (for wCard("Neighborhood"))

	LONGINT				GDBSCAN_rHoodCount;								// count calls to _retrieveNeighborhood(..)
	LONGINT				GDBSCAN_NoOfClusters;                    	// number of clusters found
   LONGINT				GDBSCAN_totalScanned;     						// total of objects scanned
   LONGINT				GDBSCAN_totalInClusters;						// total assigned to clusters
};

// ********** OPTIONAL HOOK FUNCTIONS ******************************************
template <class DATA>
void GDBSCAN_database<DATA>::GDBSCAN_prepareDatabase()		  	// optinal hook function called before clustering
{}

// ********** DEFAULT FUNCTIONS ************************************************
template <class DATA>
float GDBSCAN_database<DATA>::GDBSCAN_wCard(DATA *o)
{                                                              // slower, but universal version
	DataList result(false);
   result.insert(o);
   return GDBSCAN_wCard(&result);
}

// ********** Class GDBSCAN ****************************************************
template <class DATA>
class GDBSCAN
{
public:
  	GDBSCAN(GDBSCAN_database<DATA> *db,                         // db and
   		  clusterInfo<DATA> *clInfo) {_clInfo=clInfo;_db=db;}	// 	clInfo must be created by client

	// basic functions
	LONGINT 				GDBSCAN_doGDBSCAN();								// cluster whole database
	void					GDBSCAN_reset_single();							// reset for _expand_single
	LONGINT 				GDBSCAN_expand_single(DATA *o);	 			// expand single entry

	// virtual functions
	virtual void  		GDBSCAN_set_clusterInfo(clusterInfo<DATA> *clInfo) { _clInfo=clInfo; }
	virtual void  		GDBSCAN_set_database(GDBSCAN_database<DATA> *db) 	{ _db=db; }

	// virtual functions / optional hook functions
	virtual void		GDBSCAN_showProgressCluster() {};			// for each cluster

	// database access
	GDBSCAN_database<DATA> *_db;
   // cluster info
	clusterInfo<DATA>	*_clInfo;
	CLUSTERID 			GDBSCAN_single_clId;
   LONGINT           GDBSCAN_aClusterSize;

	// aux. functions
	void					GDBSCAN_init_clusterIds();						// can be used by app.
protected:
	bool 					GDBSCAN_expand_cluster(DATA* o, CLUSTERID clId);
	bool 					GDBSCAN_minWeight(DataList *result);	    
	// aux. functions
	CLUSTERID 			GDBSCAN_next_clusterId(CLUSTERID OldId) { return ++OldId; }
	bool					GDBSCAN_newSeedPoint(DATA *d, DataList *currentNewSeeds);
	void					GDBSCAN_updateClusterSeed(DataList *seed, DATA *o,
   															DataList *res, CLUSTERID clId);
};

// ********** GDBSCAN **********************************************************
template <class DATA>
LONGINT GDBSCAN<DATA>::GDBSCAN_doGDBSCAN()
{
   GDBSCAN_aClusterSize=0;
   _db->GDBSCAN_totalScanned=0;
   _db->GDBSCAN_totalInClusters=0;
	_db->GDBSCAN_NoOfClusters = 0;
	_clInfo->clear();
	_db->GDBSCAN_rHoodCount = 0;
   _db->GDBSCAN_prepareDatabase();
	GDBSCAN_init_clusterIds();

	CLUSTERID clId = GDBSCAN_next_clusterId(ID_NOISE);
   _db->GDBSCAN_cursor(); 										  			// open cursor for linear scan of database

	while (DATA *o=_db->GDBSCAN_fetch())								// 	process one by one (copy)
	{
		if (o->clusterId==ID_UNCLASSIFIED)
			if (_db->GDBSCAN_wCard(o)>0) 									// 	object in selection?
				if (GDBSCAN_expand_cluster(o, clId))
				{
					clId = GDBSCAN_next_clusterId(clId);				// select a new cluster-id for NEXT cluster
               GDBSCAN_showProgressCluster();
               GDBSCAN_aClusterSize=0;
				}
      delete o;
		_db->GDBSCAN_totalScanned++;
	}
	return _db->GDBSCAN_NoOfClusters;
}

template <class DATA>
void GDBSCAN<DATA>::GDBSCAN_reset_single()
{
   _db->GDBSCAN_prepareDatabase();
  	GDBSCAN_init_clusterIds();
	_db->GDBSCAN_NoOfClusters = 0;
	_clInfo->clear();
	GDBSCAN_single_clId=GDBSCAN_next_clusterId(ID_NOISE);
}

template <class DATA>
LONGINT GDBSCAN<DATA>::GDBSCAN_expand_single(DATA *o)
{
   GDBSCAN_aClusterSize=0;
	_db->GDBSCAN_rHoodCount = 0;
	if (_db->GDBSCAN_wCard(o)>0) 											// 	object in selection?
		if (GDBSCAN_expand_cluster(o, GDBSCAN_single_clId))
		{
			GDBSCAN_single_clId = GDBSCAN_next_clusterId(GDBSCAN_single_clId);
			GDBSCAN_showProgressCluster();
			return GDBSCAN_aClusterSize;
		}
	return 0;
}

// ********** EXPAND CLUSTER ***************************************************
template <class DATA>
bool GDBSCAN<DATA>::GDBSCAN_expand_cluster(DATA* o, CLUSTERID clId)
{
	DataList *result;
	_db->GDBSCAN_retrieveNeighborhood(o, &result); _db->GDBSCAN_rHoodCount++;

	if (GDBSCAN_minWeight(result))
	{	// ***** o is core point *****
		_clInfo->createTemp(clId);
		DataList *seed = new DataList();									// DataList(true) => elements deleted in destr.
		GDBSCAN_updateClusterSeed(seed, o, result, clId);
		delete result;
		while (seed->get_num() > 0)
      {
			DATA *currentPoint = seed->get_last();						// seeds.top() - new objects inserted at first() pos.
			_db->GDBSCAN_retrieveNeighborhood(currentPoint, &result); _db->GDBSCAN_rHoodCount++;
         seed->noDelete();													// make object "outside" (currentPoint is deleted later)
			seed->erase(); 													// seeds.pop()
			if (GDBSCAN_minWeight(result))
				GDBSCAN_updateClusterSeed(seed, currentPoint, result, clId);
         delete currentPoint;
			delete result;
		}
		delete seed;
		_clInfo->commitTemp();
      _db->GDBSCAN_NoOfClusters++;
		return true;
	} else
	{	// ***** o is not a core point *****
      _db->GDBSCAN_set_clusterId(o, ID_NOISE);						// mark as NOISE 						
      _clInfo->showNoise(o);
		delete result;
		return false;
	}
}

// ********** AUXIL. FUNCTIONS *************************************************
template <class DATA>
void GDBSCAN<DATA>::GDBSCAN_init_clusterIds()
{
   _db->GDBSCAN_cursor();										  			// open cursor for linear scan of database
	while (DATA *o = _db->GDBSCAN_fetch())					  			// 	process one by one (copy)
   {
	  	_db->GDBSCAN_set_clusterId(o, ID_UNCLASSIFIED);
      delete o;
	}
}

template <class DATA>
bool GDBSCAN<DATA>::GDBSCAN_newSeedPoint(DATA *d, DataList *currentNewSeeds)
{
  	if (d->clusterId != ID_UNCLASSIFIED) return false;
	for (DATA *p=currentNewSeeds->get_first(); p; p=currentNewSeeds->get_next())
   	if (d->compare(p)==0) return false;
	return true;
}

template <class DATA>
void GDBSCAN<DATA>::GDBSCAN_updateClusterSeed(DataList *seed, DATA *o,
																 DataList *result, CLUSTERID clId)
{
	DataList currentNewSeeds(false);										// list holds "outside" objects
	currentNewSeeds.insert(o);

	for (DATA *d=result->get_first(); d; d=result->get_next())
	{
		if (_db->GDBSCAN_wCard(d)>0)	  				  					// 	object in selection?
      {
         if (GDBSCAN_newSeedPoint(d, &currentNewSeeds))
         {
            GDBSCAN_aClusterSize++;
            _db->GDBSCAN_totalInClusters++;
            d->clusterId=clId;											// set new Id for seedpoints
            _db->GDBSCAN_set_clusterId(d, clId);					// 	BEFORE insert to currentNewSeeds
            _clInfo->updateTemp(d);										//			(for visualization only)
            result->noDelete();	 										// make object "outside"
            seed->insert(d);												// 	=> insert here (will be deleted here)
            currentNewSeeds.insert(d);									// list holds "outside" objects
         } else if (d->clusterId <= ID_NOISE)
         {
	      	GDBSCAN_aClusterSize++;
   	     	_db->GDBSCAN_totalInClusters++;
		      d->clusterId=clId;
            _db->GDBSCAN_set_clusterId(d, clId);
            _clInfo->updateTemp(d);
         }
		}
	}
}

template <class DATA>
bool GDBSCAN<DATA>::GDBSCAN_minWeight(DataList *result)
{	return (_db->GDBSCAN_wCard(result)>=_db->GDBSCAN_minCard); }
#endif
